// Copyright 1996-2000 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.

#import "OmniCDUnitController.h"

#import "OmniCDController.h"
#import "OmniCDGeneralPreferences.h"

#import <OmniFoundation/OmniFoundation.h>
#import <OmniSCSI/OmniSCSI.h>
#import <OmniCDDB/CDDBDatabase.h>
#import <OmniCDDB/CDDBPreferences.h>

#warning Later, we could have a 'playedTrackAttributes' for when we are in shuffle mode
static NSDictionary *currentTrackAttributes  = nil;
static NSDictionary *normalTrackAttributes   = nil;
static NSDictionary *disabledTrackAttributes = nil;

@interface OmniCDUnitController (PrivateAPI)
- (void) _setWindowTitle;
- (void) _setTrackString;
- (void) _startUpdateTimer;
- (void) _stopUpdateTimer;
- (void) _update: (NSTimer *) timer;
- (void) _setAttributes: (NSDictionary *) trackAttributes
               forTrack: (unsigned int) track;
- (void) _setSelectedTrack: (unsigned int) newTrack;
- (void) _skipToTrack: (unsigned int) trac playImmediately: (BOOL) playImmediately;
- (void) _tableOfContentsChanged: (NSNotification *) note;
- (void) _playRange: (NSRange) range;

- (void) _startQueryThreadForTableOfContents: (OSCSICDTableOfContents *) toc;
- (void) _sendQuery: (NSArray *) arguments;
- (void) _processResults: (id) results
              forRequest: (unsigned int) requestCounter;
@end

@implementation OmniCDUnitController

+ (void) initialize;
{
    NSFont *font;
    NSMutableParagraphStyle *style;
    
    [super initialize];
    if (currentTrackAttributes)
        return;

    style = [[NSMutableParagraphStyle alloc] init];
    [style setAlignment: NSCenterTextAlignment];
        
    font = [NSFont fontWithName: @"Helvetica-Bold" size: 10.0];
    currentTrackAttributes = [[NSDictionary alloc] initWithObjectsAndKeys:
        [NSColor blueColor], NSForegroundColorAttributeName,
        font, NSFontAttributeName,
        style, NSParagraphStyleAttributeName,
        nil];

    font = [NSFont fontWithName: @"Helvetica" size: 10.0];
    normalTrackAttributes = [[NSDictionary alloc] initWithObjectsAndKeys:
        [NSColor blackColor], NSForegroundColorAttributeName,
        font, NSFontAttributeName,
        style, NSParagraphStyleAttributeName,
        nil];

    disabledTrackAttributes = [[NSDictionary alloc] initWithObjectsAndKeys:
        [NSColor darkGrayColor], NSForegroundColorAttributeName,
        font, NSFontAttributeName,
        style, NSParagraphStyleAttributeName,
        nil];
}

- initWithTarget: (OSCSITarget *) target;
{
    NSString *unitDisplayName;
    
    if (!(self = [super init]))
        return nil;

    _target = [target retain];
    _audioUnit = [[target unitWithUnitNumber: 0] retain];

    NS_DURING {
        [(OSCSIUnit *)_audioUnit open];
    } NS_HANDLER {
        NSRunAlertPanel(nil, @"Error opening selected device.", @"Abort", nil, nil);
        [self release];
        return nil;
    } NS_ENDHANDLER;

    if (![NSBundle loadNibNamed: @"CDUnit" owner: self]) {
        NSRunAlertPanel(nil, @"Cannot load CDUnit.nib.", @"Abort", nil, nil, nil);
        [self release];
        return nil;
    }

    //NSLog(@"Device name = %@", [[(OSCSIUnit *)_audioUnit unitDescription] unitDisplayName]);
    
    _database = [[CDDBDatabase alloc] init];
    
    _supportsIdentification = [_audioUnit conformsToProtocol: @protocol(OSCSICDRomAudioIdentification)];
    _supportsVolumeControl  = [_audioUnit conformsToProtocol: @protocol(OSCSICDRomAudioVolumeControl)];
    _supportsScan           = [_audioUnit conformsToProtocol: @protocol(OSCSICDRomAudioScan)];

    unitDisplayName = [[(OSCSIUnit *)_audioUnit unitDescription] unitDisplayName];
    [window setTitle: unitDisplayName];

    [[NSNotificationCenter defaultCenter] addObserver: self
                                             selector: @selector(_tableOfContentsChanged:)
                                                 name: OSCSICDAudioUnitTableOfContentsChangedNotification
                                               object: _audioUnit];

    [window setFrameUsingName:unitDisplayName];
    [window setFrameAutosaveName: unitDisplayName];
            
    return self;
}

- (void) awakeFromNib;
{
    [[playButton cell] setShowsStateBy: NSContentsCellMask];
    [[pauseButton cell] setShowsStateBy: NSContentsCellMask];
    [self _setSelectedTrack: -1];
}

- (void) dealloc;
{
    [[NSNotificationCenter defaultCenter] removeObserver: self];
    [self _stopUpdateTimer];
    [_audioUnit release];
    [_target release];
    [_mountPoint release];
    [_unitConnection release];
    [(id)_database release];
    [_discInfo release];
    [super dealloc];
}

- (void) mediaMountedAtPath: (NSString *) mountPoint;
{
    [_mountPoint release];
    _mountPoint = [mountPoint copy];

    [window orderFront: nil];

    // Make sure _tableOfContentsChanged: gets called
    [_audioUnit updateStatus];

    // Allow for any weirdness that might happen
    if ([_audioUnit mediaPresent]) {
        if ([[OFUserDefaults sharedUserDefaults] boolForKey: OCDAutoPlayOnDiscInsert])
            [self play: nil];
    }
}

- (IBAction) play: (id) sender;
{
    OSCSICDTableOfContents *toc;
    OSCSILogicalBlockAddress     start, end;
    
    toc = [_audioUnit tableOfContents];

    // This doesn't seem to work (at least if we aren't already playing)
    //start = OSCSI_USE_CURRENT_PLAY_POSITION;
    if (_changeTrackOnPlay) {
        start = [toc startingLBAForTrackNumber: _selectedTrack];
        _changeTrackOnPlay = NO;
    } else
        start = [toc startingLBAForTrackNumber: 1];
    end = [toc startingLBAForTrackNumber: OSCSI_LEAD_OUT_TRACK];

    [self _playRange: NSMakeRange(start, end - start)];
}

- (IBAction) pause: (id) sender;
{
    BOOL isPaused;

    [_audioUnit updateStatus];
    isPaused = [_audioUnit currentStatus] == OSCSICDAudioPausedStatus;

    if (_changeTrackOnPlay && isPaused) {
        // Start playing from the new selected track.
        [self _skipToTrack: _selectedTrack playImmediately: YES];
        _changeTrackOnPlay = NO;
    }
    // It is conceivable that some players might not unpause when we send a play
    // command.  So, we'll make sure that we unpause anyway.
    [_audioUnit setAudioIsPaused: !isPaused];
}

- (IBAction) stopPlaying: (id) sender;
{
    [_audioUnit stopPlaying];
    [self _update: nil];
}

- (IBAction) forward: (id) sender;
{
    OSCSICDTableOfContents *toc;
    unsigned int                 newTrack;

    toc = [_audioUnit tableOfContents];
    newTrack = _selectedTrack + 1;
    if (newTrack > [toc numberOfTracks])
        newTrack = 1;
    [self _skipToTrack: newTrack playImmediately: NO];
}

- (IBAction) backward: (id) sender;
{
    OSCSICDTableOfContents *toc;
    unsigned int                 newTrack;

    toc = [_audioUnit tableOfContents];
    newTrack = _selectedTrack - 1;
    if (newTrack < 1)
        newTrack = [toc numberOfTracks];
    [self _skipToTrack: newTrack playImmediately: NO];
}

- (IBAction) setTrack: (id) sender;
{
    unsigned int                 newTrack;

    newTrack = [[sender selectedCell] tag] + 1;
    [self _skipToTrack: newTrack playImmediately: NO];
}

- (IBAction) setVolume: (id) sender;
{
    if (_supportsVolumeControl) {
        float percentage = [sender floatValue];
        
        [(id <OSCSICDRomAudioVolumeControl>) _audioUnit setLeftVolumePercentage: percentage
                                                          rightVolumePercentage: percentage];
        [self _update: nil];
    }
}

- (IBAction) eject: (id) sender;
{
#if 0 // Since we don't really mount the device, we can't ask workspace to eject it.
    if (![[NSWorkspace sharedWorkspace] unmountAndEjectDeviceAtPath: _mountPoint]) {
        NSLog(@"Error unmounting device via Workspace.  Will attempt SCSI eject.");
#else
    if (1) {
#endif
        // We'll do it the hard way then
        [(OSCSIUnit *)_audioUnit setAllowsMediaRemoval: YES];
        [(OSCSIUnit *)_audioUnit eject];
    }

    [_mountPoint release];
    _mountPoint = nil;    
}

- (IBAction) queryDatabase: (id) sender;
{
    if (_mediaPresent) {
        OSCSICDTableOfContents *toc;

        if ((toc = [_audioUnit tableOfContents]))
            [self _startQueryThreadForTableOfContents: toc];
    }
}

- (IBAction) changeTimeDisplayFormat: (id) sender;
{
    _timeDisplayFormat = _timeDisplayFormat + 1;
    if (_timeDisplayFormat == OmniCDLastTimeDisplayFormat) {
        _timeDisplayFormat = 0;

        switch (_timeDisplayFormat) {
            case OmniCDElapsedSinceStartOfCurrentTrackTimeDisplayFormat:
                [currentTimeButton setToolTip: @"Elapsed Time In Current Track"];
                break;
            case OmniCDRemainingInCurrentTrackTimeDisplayFormat:
                [currentTimeButton setToolTip: @"Remaining Time In Current Track"];
                break;
            case OmniCDTotalInCurrentTrackTimeDisplayFormat:
                [currentTimeButton setToolTip: @"Total Time In Current Track"];
                break;
            case OmniCDElapsedSinceStartOfDiscTimeFormat:
                [currentTimeButton setToolTip: @"Elapsed Time In Disc"];
                break;
            case OmniCDRemainingOnDiscTimeDisplayFormat:
                [currentTimeButton setToolTip: @"Remaining Time In Disc"];
                break;
            case OmniCDTotalOnDiscTimeDisplayFormat:
                [currentTimeButton setToolTip: @"Total Time In Disc"];
                break;
            default:
                break;
        }
    }
}

//
// NSWindow delegate
//

- (void)windowWillClose:(NSNotification *)notification;
{
    OmniCDController *controller;
    NSString         *deviceName;

    if (_mediaPresent)
        [self eject: nil];

    controller = (OmniCDController *)[NSApp delegate];
    deviceName = [[(OSCSIUnit *)_audioUnit target] rawDeviceName];

    [controller unitControllerClosedForTarget: _target];
}

@end


@implementation OmniCDUnitController (PrivateAPI)

- (void) _setWindowTitle;
{
    NSString *windowTitle;
    NSString *artist, *title;

    if (_mediaPresent) {
        if (!(artist = [_discInfo objectForKey: CDDBDiscArtistKey]))
            artist = @"Unknown";

        if (!(title  = [_discInfo objectForKey: CDDBDiscTitleKey]))
            title = @"Unknown";

#warning This should probably be a preference.  It is often the case that "Artist - Title" is too long
        //windowTitle = [NSString stringWithFormat: @"%@ - %@", artist, title];
        windowTitle = title;
    } else
        windowTitle = @"No Disc Loaded";

    [window setTitle: windowTitle];
}

- (void) _setTrackString;
{
    NSString                    *title;
    NSArray                     *tracks;
    OSCSICDTableOfContents *toc = nil;
    NSString                    *trackString;

    if (_mediaPresent) {
        tracks = [_discInfo objectForKey: CDDBTrackInfoKey];
        if ([tracks count] <= _selectedTrack - 1)
            // Could be a nil _discInfo or bad data
            title = [NSString stringWithFormat: @"Track #%d", _selectedTrack < 1 ? 1 : _selectedTrack];
        else
            title = [[tracks objectAtIndex: _selectedTrack - 1] objectForKey: CDDBTrackTitleKey];

        toc = [_audioUnit tableOfContents];

#warning This looks better in the common case, but we really need a good way to show how many tracks there are total (since the track matrix might not be smaller than the actual number of tracks).
        //trackString = [NSString stringWithFormat: @"[%d/%d] %@", _selectedTrack, [toc numberOfTracks], title];
        trackString = title;
    } else
        trackString = @"";

    [currentTrackTextField setStringValue: trackString];
}

- (void) _startUpdateTimer;
{
    if (!_updateTimer)
        _updateTimer = [[NSTimer scheduledTimerWithTimeInterval: 0.5
                                                        target: self
                                                      selector: @selector(_update:)
                                                      userInfo: nil
                                                       repeats: YES] retain];
}

- (void) _stopUpdateTimer;
{
    [_updateTimer invalidate];
    [_updateTimer release];
    _updateTimer = nil;
}

- (void) _update: (NSTimer *) timer;
{
    BOOL                         playing;
    OSCSICDAudioStatus           status;
    OSCSICDTableOfContents *toc = nil;
    unsigned int                 currentTrack;
    NSString                    *timeString;
    OSCSILogicalBlockAddress     lba, end;

    if (_mediaPresent) {
        [_audioUnit updateStatus];
        // Make sure the table of contents gets loaded
        toc = [_audioUnit tableOfContents];
    }


    [unloadButton setEnabled: _mediaPresent];
    [forwardButton setEnabled: _mediaPresent];
    [backwardButton setEnabled: _mediaPresent];
    [volumeSlider setEnabled: _mediaPresent];
    
    status = [_audioUnit currentStatus];
    playing = (status == OSCSICDAudioPlayingStatus) || (status == OSCSICDAudioPausedStatus);
    
    [pauseButton setState: (status == OSCSICDAudioPausedStatus)];
    [pauseButton setEnabled: playing];

    [playButton setEnabled: _mediaPresent];
    [playButton setState: playing];
    
    [stopButton setEnabled: playing];


    if (_mediaPresent) {
        if (status == OSCSICDAudioIdleStatus)
            _currentLBA = 0;
        else if (status != OSCSICDAudioPausedStatus)
            _currentLBA = [_audioUnit currentLBA];
    } else
        _currentLBA = 0;
    if ((int)_currentLBA < 0)
        _currentLBA = 0;
    
    end = [toc startingLBAForTrackNumber: OSCSI_LEAD_OUT_TRACK];
    if (_currentLBA >= end)
        // We've hit the end of the disc
        currentTrack = [toc numberOfTracks];
    else
        currentTrack = [toc trackContainingLBA: _currentLBA];
    if (status == OSCSICDAudioPlayingStatus) {
        [self _setSelectedTrack: currentTrack];
    }

    if (_mediaPresent) {
        OSCSIMinuteSecondsFrames     msf;

        switch (_timeDisplayFormat) {
            case OmniCDElapsedSinceStartOfCurrentTrackTimeDisplayFormat:
                lba = _currentLBA - [toc startingLBAForTrackNumber: currentTrack];
                break;
            case OmniCDRemainingInCurrentTrackTimeDisplayFormat:
                lba = [toc endingLBAForTrackNumber: currentTrack] - _currentLBA;
                break;
            case OmniCDTotalInCurrentTrackTimeDisplayFormat:
                lba = [toc endingLBAForTrackNumber: currentTrack] - [toc startingLBAForTrackNumber: currentTrack];
                break;
            case OmniCDElapsedSinceStartOfDiscTimeFormat:
                lba = _currentLBA;
                break;
            case OmniCDRemainingOnDiscTimeDisplayFormat:
                lba = end - _currentLBA;
                break;
            case OmniCDTotalOnDiscTimeDisplayFormat:
                lba = end;
                break;
            default:
                lba = 0;
                break;
        }

        msf = OSCSIRelativeLogicalBlockAddressToMinutesSecondsFrames(lba);
        timeString = [NSString stringWithFormat: @"%02d:%02d", msf.minutes, msf.seconds];
    } else
        timeString = @"";
    [currentTimeButton setTitle: timeString];

    // For now, we are assuming they are the same
    if (_supportsVolumeControl) {
        float volumePercentage;

        volumePercentage = [(id <OSCSICDRomAudioVolumeControl>)_audioUnit leftVolumePercentage];

        // "Our volume goes up to *11* !"
        //[currentVolumeTextField setStringValue: [NSString stringWithFormat: @"%d", (int)(volumePercentage * 11)]];
        [volumeSlider setFloatValue: volumePercentage];
    }
}

- (void) _setAttributes: (NSDictionary *) trackAttributes
               forTrack: (unsigned int) track;
{
    NSButtonCell       *cell;
    NSAttributedString *string;

    cell = [trackMatrix cellWithTag: track - 1];
    string = [[NSAttributedString alloc] initWithString: [NSString stringWithFormat: @"%d", track]
                                             attributes: trackAttributes];
    [cell setAttributedTitle: string];

}

- (void) _setSelectedTrack: (unsigned int) newTrack;
{

    if (_selectedTrack == newTrack)
        return;

    [self _setAttributes: normalTrackAttributes forTrack: _selectedTrack];
    [self _setAttributes: currentTrackAttributes forTrack: newTrack];

    _selectedTrack = newTrack;
    [self _setTrackString];
}

- (void) _skipToTrack: (unsigned int) track playImmediately: (BOOL) playImmediately;
{
    OSCSICDAudioStatus           status;

    [self _setSelectedTrack: track];

    status = [_audioUnit currentStatus];

    if (playImmediately || status == OSCSICDAudioPlayingStatus) {
        OSCSILogicalBlockAddress     start, end;
        OSCSICDTableOfContents *toc;

        // Start playing at the beginning of the next track, continuing until the
        // end of the disc.  If we are on the last track, start on the first track.
        toc   = [_audioUnit tableOfContents];
        start = [toc startingLBAForTrackNumber: _selectedTrack];
        end   = [toc startingLBAForTrackNumber: OSCSI_LEAD_OUT_TRACK];

        [self _playRange: NSMakeRange(start, end - start)];
    } else if (status == OSCSICDAudioPausedStatus || OSCSICDAudioIdleStatus) {
        // In this case, we want to just remember the fact that when we get unpaused
        // or start playing , we should start playing at the requested location.
        _changeTrackOnPlay = YES;
    } else {
        // Huh.
        return;
    }
}

- (void) _tableOfContentsChanged: (NSNotification *) note;
{
    NSArray                     *cells;
    unsigned int                 cellIndex, numberOfTracks;
    OSCSICDTableOfContents *toc;

    toc = [[note object] tableOfContents];

    [_discInfo release];
    _discInfo = nil;
    if (toc) {
        _mediaPresent = YES;
        [self _startUpdateTimer];

        if ([[OFUserDefaults sharedUserDefaults] boolForKey: CDDBServerAutoQueryDefaultKey])
            [self queryDatabase: nil];
        else {
            NSString *unitDisplayName;

            // Can't just leave the 'No Disc Loaded' there
            unitDisplayName = [[(OSCSIUnit *)_audioUnit unitDescription] unitDisplayName];
            [window setTitle: unitDisplayName];
        }
    } else {
        _mediaPresent = NO;
        [self _stopUpdateTimer];
        
        [self _setWindowTitle];
        [self _setTrackString];
    }

    // Disable cells for tracks that don't exist on this disk
    numberOfTracks = [toc numberOfTracks];
    
    cells = [trackMatrix cells];
    cellIndex = [cells count];
    while (cellIndex--) {
        NSCell *cell;
        NSDictionary *attributes;
        

        cell = [cells objectAtIndex: cellIndex];
        if ([cell tag] < numberOfTracks) {
            [cell setEnabled: YES];
            attributes = normalTrackAttributes;
        } else {
            [cell setEnabled: NO];
            attributes = disabledTrackAttributes;
        }

        // If you have set an attributed title, disabling an NSButtonCell won't gray it out
        [self _setAttributes: attributes forTrack: cellIndex + 1];
    }

    [trackMatrix setNeedsDisplay: YES];
    if (numberOfTracks) {
        _selectedTrack = -1;
        [self _setSelectedTrack: 1];
    }
}

- (void) _playRange: (NSRange) range;
{
    [_audioUnit startPlayingLBARange: range];

    // The player will start doing what we ask, but it won't do it immediately.
    // Also, seeking on a CD player is not exact.  If we aren't careful, we
    // could briefly be in a state that the player claims to be playing just
    // before the point that we requested.  If this is across a track boundary
    // and we get an _update: message in that time, we'll briefly display
    // 'incorrect' information.  To avoid that, we'll stop the update timer
    // and then restart it after a little bit.
    [self _stopUpdateTimer];
    [self performSelector: @selector(_startUpdateTimer) withObject: nil afterDelay: 1.0];
}

- (void) _startQueryThreadForTableOfContents: (OSCSICDTableOfContents *) toc;
{
    NSArray      *arguments;
    NSPort       *sendPort, *receivePort;

    // Since we can have multiple units active at one time, we need to have each unit have
    // its own connection.  We could have used +defaultConnection, but then we couldn't
    // set multiple root objects, one per unit.
    if (!_unitConnection) {
        sendPort = [NSPort port];
        receivePort = [NSPort port];
        _unitConnection = [[NSConnection alloc] initWithReceivePort:receivePort sendPort:sendPort];
        [_unitConnection setRootObject: self];
    } else {
        sendPort = [_unitConnection sendPort];
        receivePort = [_unitConnection receivePort];
    }
    
#warning Make sure that either DO is retaining us while the thread is running, or that we otherwise keep ourselves from getting deallocated while a query is in progress.

    // We'll send a counter along with the request.  When the response comes back (if it ever does), we'll
    // compare the returned counter with our current counter.  If they differ, then we must have ejected
    // the disc in the mean time and started another request.  We could use the TOC, but if we pass it
    // back across DO, we'll actually get a proxy.  Also, it isn't guaranteed to be unique.  The old TOC
    // could have gotten deallocated and a new TOC allocated in exactly the same spot in memory, leading
    // to a false positive.
    _requestCounter++;
    arguments = [NSArray arrayWithObjects: sendPort, receivePort, toc, [NSNumber numberWithInt: _requestCounter], nil];

#warning This is not as safe as it should be.  If we get an invocation of this message while we still have one pending, our _database ivar will get sent a -sendQuery: message while one is already in progress.
    [NSThread detachNewThreadSelector: @selector(_sendQuery:) toTarget: self withObject: arguments];
}

- (void) _sendQuery: (NSArray *) arguments;
{
    NSAutoreleasePool           *pool;
    NSConnection                *replyConnection;
    id                           results;
    OmniCDUnitController        *selfProxy;
    OSCSICDTableOfContents *toc;
    
    //
    // This happens in a separate thread, so we shouldn't touch anything
    // but the database unless we do so through Distributed Objects.
    //

    pool = [[NSAutoreleasePool alloc] init];

    toc = [arguments objectAtIndex: 2];
    results = [_database queryDatabase: toc];

    replyConnection = [NSConnection connectionWithReceivePort: [arguments objectAtIndex: 0]
                                                     sendPort: [arguments objectAtIndex: 1]];
    // Normally, we'd set a protocol on this proxy, but since the number of
    // message will be *very* low, we won't bother.
    selfProxy = (OmniCDUnitController *)[replyConnection rootProxy];

    [selfProxy _processResults: results
                    forRequest: [[arguments objectAtIndex: 3] unsignedIntValue]];

    [pool release];
}

- (void) _processResults: (id) results
              forRequest: (unsigned int) requestCounter;
{
    //
    // We're in the main thread now.
    //
    if (!_mediaPresent || requestCounter != _requestCounter) {
        // This could be an old query that was submitted when we had a different disc
        // in the player.  Heck, we might not have even put in a new disc.  Ignore it.
        return;
    }

    if ([results isKindOfClass: [NSException class]]) {
        NSRunAlertPanel(@"CD Database Error", [results reason], @"Ok", nil, nil);
        return;
    }
    
    // Later, we'll want to pop up a panel if there are multiple
    // results.  For now, we'll just log a message.
    if ([results count] > 1) {
        NSLog(@"Database query returned %d results.  Can currently only handle singular result sets.  Using first result", [results count]);
    }

    [_discInfo release];
    _discInfo = [[results objectAtIndex: 0] retain];

    [self _setWindowTitle];
    [self _setTrackString];
}


@end
